use std::{
    collections::{BTreeMap, HashMap}, iter::Map, net::IpAddr, sync::RwLock
};

use super::util::Address;

/* -------------------------------------------------------------------------- */
/*                              Instance Storage                              */
/* -------------------------------------------------------------------------- */

#[derive(Debug)]
pub enum InstanceStorageError {
    NotFound,
    Full
}

pub trait Instance {
    type Id;
    fn id(&self) -> Self::Id;
    fn ip(&self) -> IpAddr;
}

pub trait InstanceStorage<Item: Instance>: IntoIterator<Item = Item> {
    // Create

    // Request
    fn get_by_address(&self, address: &Address) -> Result<&Item, InstanceStorageError>;
    fn get_by_address_mut(
        &mut self,
        address: &Address,
    ) -> Result<&mut Item, InstanceStorageError>;
    fn get_by_id(&self, id: Item::Id) -> Result<&Item, InstanceStorageError>;
    fn get_by_id_mut(&mut self, id: Item::Id) -> Result<&mut Item, InstanceStorageError>;
    // Delete
    fn remove_by_address(&mut self, address: &Address) -> Result<(), InstanceStorageError>;
    fn remove_by_id(&mut self, id: Item::Id) -> Result<(), InstanceStorageError>;
}

/* ------------------------- Default Implementation ------------------------- */
pub struct BasicInstance {
    id: u32,
    ip: IpAddr
}

impl Instance for BasicInstance {
    type Id = u32;
    fn id(&self) -> Self::Id {
        self.id
    }
    fn ip(&self) -> IpAddr {
        self.ip
    }
}

pub struct BasicInstanceStorage {
    max_instances: u32,
    instances: BTreeMap<Address, BasicInstance>,
}

impl BasicInstanceStorage {
    fn new(max_instances: u32) -> Self {
        Self {
            instances: BTreeMap::new(),
            max_instances
        }
    }
}

impl InstanceStorage<BasicInstance> for BasicInstanceStorage {
    fn get_by_address(&self, address: &Address) -> Result<&BasicInstance, InstanceStorageError> {
        match self.instances.get(address) {
            Some(v) => {
                Ok(v)
            },
            None => Err(InstanceStorageError::NotFound)
        }
    }

    fn get_by_address_mut(
        &mut self,
        address: &Address,
    ) -> Result<&mut BasicInstance, InstanceStorageError> {
        match self.instances.get_mut(address) {
            Some(v) => {
                Ok(v)
            },
            None => Err(InstanceStorageError::NotFound)
        }
    }

    fn get_by_id(&self, id: <BasicInstance as Instance>::Id) -> Result<&BasicInstance, InstanceStorageError> {
        let search_result  = self.instances
            .iter()
            .find(|v| v.1.id == id);
        match search_result   {
            Some(v)   =>  {
                Ok(v.1)
            },
            None   =>  {
                Err(InstanceStorageError::NotFound)
            }
        }
    }

    fn get_by_id_mut(&mut self, id: <BasicInstance as Instance>::Id) -> Result<&mut BasicInstance, InstanceStorageError> {
        let search_result  = self.instances
            .iter_mut()
            .find(|v| v.1.id == id);
        match search_result   {
            Some(v)   =>  {
                Ok(v.1)
            },
            None   =>  {
                Err(InstanceStorageError::NotFound)
            }
        }
    }

    fn remove_by_address(&mut self, address: &Address) -> Result<(), InstanceStorageError> {
        match self.instances.remove(address) {
            Some(_) => {
                Ok(())
            },
            None => Err(InstanceStorageError::NotFound)
        }
    }

    fn remove_by_id(&mut self, id: <BasicInstance as Instance>::Id) -> Result<(), InstanceStorageError> {
        let original_len = self.instances.len();
        self.instances.retain(|_, v| v.id != id);
    
        if self.instances.len() == original_len {
            return Err(InstanceStorageError::NotFound);
        }
    
        Ok(())
    }    
}

impl IntoIterator for BasicInstanceStorage {
    type Item = BasicInstance;
    type IntoIter = std::vec::IntoIter<BasicInstance>;

    fn into_iter(self) -> Self::IntoIter {
        self.instances
            .into_iter()
            .map(|(_address, instance)| instance)
            .collect::<Vec<BasicInstance>>()
            .into_iter()
    }
}

#[cfg(test)]
mod basic_instance_test {
    use std::net::{IpAddr, Ipv4Addr};

    use macaddr::{MacAddr, MacAddr6};

    use super::*;

    fn _create_sample_address(id: u8) -> Address {
        let ip: IpAddr = IpAddr::V4(Ipv4Addr::new(127, 0, 0, id));
        let mac: MacAddr  = MacAddr::V6(MacAddr6::new(id, id, id, id, id, id));

        Address::new(ip, mac)
    }

    #[test]
    fn test_get_by_address_found() {
        let mut storage = BasicInstanceStorage::new(10);
        let instance = BasicInstance { id: 42, ip: IpAddr::V4(Ipv4Addr::LOCALHOST) };
        let addr = _create_sample_address(1);
        storage.instances.insert(addr.clone(), instance);
        let fetched = storage.get_by_address(&addr).unwrap();
        assert_eq!(fetched.id, 42);
    }

    #[test]
    fn test_get_by_address_not_found() {
        let storage = BasicInstanceStorage::new(10);
        let addr = _create_sample_address(1);
        let result = storage.get_by_address(&addr);
        assert!(matches!(result, Err(InstanceStorageError::NotFound)));
    }

    #[test]
    fn test_get_by_address_mut() {
        let mut storage = BasicInstanceStorage::new(10);
        let addr = _create_sample_address(1);
        let instance = BasicInstance { id: 42, ip: IpAddr::V4(Ipv4Addr::LOCALHOST) };
        storage.instances.insert(addr.clone(), instance);
        {
            let fetched_mut = storage.get_by_address_mut(&addr).unwrap();
            fetched_mut.id = 100;
        }
        let fetched = storage.get_by_address(&addr).unwrap();
        assert_eq!(fetched.id, 100);
    }

    #[test]
    fn test_get_by_id_found() {
        let mut storage = BasicInstanceStorage::new(10);
        let addr = _create_sample_address(1);
        let instance = BasicInstance { id: 42, ip: IpAddr::V4(Ipv4Addr::LOCALHOST) };
        storage.instances.insert(addr, instance);
        let fetched = storage.get_by_id(42).unwrap();
        assert_eq!(fetched.id, 42);
    }

    #[test]
    fn test_get_by_id_not_found() {
        let storage = BasicInstanceStorage::new(10);
        let result = storage.get_by_id(42);
        assert!(matches!(result, Err(InstanceStorageError::NotFound)));
    }

    #[test]
    fn test_get_by_id_mut() {
        let mut storage = BasicInstanceStorage::new(10);
        let addr = _create_sample_address(1);
        let instance = BasicInstance { id: 42, ip: IpAddr::V4(Ipv4Addr::LOCALHOST) };
        storage.instances.insert(addr, instance);
        {
            let fetched_mut = storage.get_by_id_mut(42);
            assert!(fetched_mut.is_ok());
            let fetched_mut = fetched_mut.unwrap();
            fetched_mut.id = 99;
        }
        let fetched = storage.get_by_id(42);
        assert!(fetched.is_err());
    }

    #[test]
    fn test_remove_by_address() {
        let mut storage = BasicInstanceStorage::new(10);
        let addr = _create_sample_address(1);
        storage.instances.insert(addr.clone(), BasicInstance { id: 42, ip: IpAddr::V4(Ipv4Addr::LOCALHOST) });
        assert!(storage.remove_by_address(&addr).is_ok());
        assert!(storage.instances.is_empty());
    }

    #[test]
    fn test_remove_by_address_not_found() {
        let mut storage = BasicInstanceStorage::new(10);
        let addr = _create_sample_address(1);
        let result = storage.remove_by_address(&addr);
        assert!(matches!(result, Err(InstanceStorageError::NotFound)));
    }

    #[test]
    fn test_remove_by_id() {
        let mut storage = BasicInstanceStorage::new(10);
        let addr = _create_sample_address(1);
        storage.instances.insert(addr, BasicInstance { id: 42, ip: IpAddr::V4(Ipv4Addr::LOCALHOST) });
        assert!(storage.remove_by_id(42).is_ok());
        assert!(storage.instances.is_empty());
    }

    #[test]
    fn test_remove_by_id_not_found() {
        let mut storage = BasicInstanceStorage::new(10);
        let result = storage.remove_by_id(42);
        assert!(matches!(result, Err(InstanceStorageError::NotFound)));
    }

    #[test]
    fn test_into_iterator() {
        let mut storage = BasicInstanceStorage::new(10);
        storage.instances.insert(_create_sample_address(1), BasicInstance { id: 42, ip: IpAddr::V4(Ipv4Addr::LOCALHOST) });
        storage.instances.insert(_create_sample_address(2), BasicInstance { id: 43, ip: IpAddr::V4(Ipv4Addr::LOCALHOST) });
        let instances: Vec<_> = storage.into_iter().collect();
        assert_eq!(instances.len(), 2);
        let ids: Vec<u32> = instances.into_iter().map(|inst| inst.id).collect();
        assert!(ids.contains(&42));
        assert!(ids.contains(&43));
    }
}

/* -------------------------------------------------------------------------- */
/*                              Attribute System                              */
/* -------------------------------------------------------------------------- */

// TODO: Optimize the heck out of this - This is effectively a placeholder
// It will likely be *very* space inefficient.

/* -------------------------------- Metadata -------------------------------- */
#[repr(u8)]
#[derive(PartialEq)]
enum AttributeType {
    Float,
    Integer,
    Boolean,
    String,
}

enum AttributeValue {
    Float(f32),
    Integer(i32),
    Boolean(bool),
    String(String),
}

impl AttributeValue {
    fn get_type(&self) -> AttributeType {
        match self {
            Float => AttributeType::Float,
            Integer => AttributeType::Integer,
            Boolean => AttributeType::Boolean,
            String => AttributeType::String,
        }
    }
}

enum AttributeError {
    InvalidType(AttributeType),
    InvalidKey(String)
}

/* --------------------------------- Schema --------------------------------- */

// TODO: Builder pattern to create this
pub struct AttributeSchema {
    attribute_schema: Map<String, AttributeSchemaElement>,
}

pub struct AttributeSchemaElement {
    default_type_and_value: AttributeValue,
}

/* ---------------------------------- List ---------------------------------- */

struct AttributeListElement {
    value: AttributeValue,
}

pub struct AttributeList<'a> {
    schema: &'a AttributeSchema,
    data: RwLock<HashMap<String, AttributeListElement>>,
}

impl<'a> AttributeList<'a> {
    fn new(schema: &'a AttributeSchema) -> Self {
        Self {
            schema,
            // TODO: Populate with initial values
            data: RwLock::new(HashMap::new()),
        }
    }

    fn set_attribute(&mut self, key: &str, value: AttributeValue) -> Result<(), AttributeError> {
        // ! Naive unwrap is present
        let mut write = self.data.write().unwrap();
        {
            let element_read = &write[key];
            if value.get_type() != element_read.value.get_type() {
                return Err(AttributeError::InvalidType(value.get_type()));
            }
        }
        {
            // ! Double query
            let element = write.get_mut(key);
            match element {
                Some(v) => {
                    v.value = value;
                },
                None => {
                    // TODO: Verify that this is the best thing to do
                    return Err(AttributeError::InvalidKey(key.to_string()))
                }
            }
        }

        Ok(())
    }
}
